@@ -45,4 +45,18 @@ def check_voxels(voxels):
     d3 = (block - voxels[:-1, :-1, 1:]).reshape(-1)
 
     max_distance = max(np.max(d1), np.max(d2), np.max(d3))
-    return max_distance < 2.0 / voxels.shape[0] * 3**0.5 * 1.1
\ No newline at end of file
+    return max_distance < 2.0 / voxels.shape[0] * 3**0.5 * 1.1
+
+def sample_uniform_points_in_unit_sphere(amount):
+    unit_sphere_points = np.random.uniform(-1, 1, size=(amount * 2 + 20, 3))
+    unit_sphere_points = unit_sphere_points[np.linalg.norm(unit_sphere_points, axis=1) < 1]
+
+    points_available = unit_sphere_points.shape[0]
+    if points_available < amount:
+        # This is a fallback for the rare case that too few points are inside the unit sphere
+        result = np.zeros((amount, 3))
+        result[:points_available, :] = unit_sphere_points
+        result[points_available:, :] = sample_uniform_points_in_unit_sphere(amount - points_available)
+        return result
+    else:
+        return unit_sphere_points[:amount, :]
